"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Hooks = void 0;
const tryCatchFinally_1 = require("./tryCatchFinally");
/** @inheritDoc {@link ITryCatchFinallyHooksCollection} */
class Hooks {
    constructor() {
        this.hooks = [];
        this._currentContext = undefined;
    }
    add(hook) {
        if (hook instanceof Function)
            return this.add({ onTry: hook });
        this.hooks.push(hook);
        return this;
    }
    //wrap(args?: DecoratorArgsOf<FullHookContext>): <TFunc extends WrappableFunction>(func:TFunc)=>TFunc // conflicts with decor
    wrap() {
        var _a;
        const _this = this;
        const beforeHooksTry = this.beforeHooksTry.bind(this);
        const afterHooksTry = (_a = this.afterHooksTry) === null || _a === void 0 ? void 0 : _a.bind(this);
        const [args, func] = arguments.length >= 2
            ? [arguments[0], arguments[1]]
            : arguments[0] instanceof Function ? [undefined, arguments[0]]
                : [arguments[0], undefined];
        if (!func)
            return (func) => this.wrap(args, func);
        return (0, tryCatchFinally_1.createTryCatchFinally)(func, {
            onTry(ctxTry) {
                if (args)
                    ctxTry.args = args;
                const hooksSorted = _this.hooks;
                const bht = beforeHooksTry(ctxTry);
                let onTryHooks = hooksSorted.map(hook => hook.onTry(ctxTry)).filter(h => h);
                onTryHooks = onTryHooks.reduce(([first, last], h) => ('lastInQueue' in h && h.lastInQueue ? last.push(h) : first.push(h), [first, last]), [[], []]).flat();
                return {
                    onFinally(ctxFinally) {
                        var _a, _b;
                        for (const hr of onTryHooks) {
                            (_a = hr.onFinally) === null || _a === void 0 ? void 0 : _a.call(hr, ctxFinally);
                        }
                        (_b = bht.onFinally) === null || _b === void 0 ? void 0 : _b.call(bht);
                    },
                    onCatch(ctxCatch) {
                        var _a, _b;
                        for (const hr of onTryHooks) {
                            (_a = hr.onCatch) === null || _a === void 0 ? void 0 : _a.call(hr, ctxCatch);
                        }
                        (_b = bht.onCatch) === null || _b === void 0 ? void 0 : _b.call(bht);
                    }
                };
            },
        });
    }
    /**
     * @param args
     * @returns
     */
    decor(args) {
        return ((_target, _propertyKey, descriptor) => {
            if (arguments.length == 1 && typeof _target === 'function')
                return this.wrap(args, _target);
            const originalMethod = descriptor.value;
            descriptor.value = this.wrap(args, originalMethod);
        });
    }
    scope(args, scope) {
        return this.wrap(args, scope)();
    }
    beforeHooksTry(ctx) {
        const _this = this;
        const prevContext = _this._currentContext;
        this._currentContext = ctx;
        return {
            onFinally() {
                _this._currentContext = prevContext;
            },
        };
    }
    /**
     * Only safe to use in sync functions or in async functions before any awaited code.
     * Otherwise, the context may be changed by another async function - in that case use callStack hook instead
     */
    get context() {
        return this._currentContext;
    }
    create() {
        const hookFunc = function () {
            const hooks = hookFunc.hooks;
            if (arguments[0] instanceof Function || arguments[1] instanceof Function)
                return hooks.wrap.apply(hooks, arguments);
            //decor or wrap
            return hooks.decor.apply(hooks, arguments);
        };
        hookFunc.hooks = this;
        hookFunc.scope = this.scope.bind(this);
        Object.defineProperty(hookFunc, 'context', { get: () => hookFunc.hooks.context });
        return hookFunc; //IHooksFunc<Hooks<FullHookContext>>
    }
}
exports.Hooks = Hooks;
