export type WrappableFunction = (this: any, ...args: any[]) => any;
export type FunctionContext<TFunc extends WrappableFunction = WrappableFunction> = {
    func: TFunc;
    funcWrapper: TFunc;
    funcArgs: Parameters<TFunc>;
    funcThis?: ThisParameterType<TFunc>;
    funcOutcome?: FunctionExecutionOutcome<TFunc>;
};
export type FunctionInterceptors<TFunc extends WrappableFunction = WrappableFunction, TContext extends {} = {}> = {
    onTry: (ctx: FunctionContext<TFunc> & TContext) => {
        context?: TContext;
        onCatch?: (ctx: TContext & FunctionContext<TFunc>) => void;
        onFinally?: (ctx: TContext & FunctionContext<TFunc>) => void;
    };
    onCatch?: (ctx: TContext & FunctionContext<TFunc>) => void;
    onFinally?: (ctx: TContext & FunctionContext<TFunc>) => void;
};
export declare function createTryCatchFinally<TFunc extends WrappableFunction = WrappableFunction, TContext extends {} = {}>(func: TFunc, interceptors: FunctionInterceptors<TFunc, TContext>): TFunc;
export type FunctionExecutionOutcome<TFunc extends WrappableFunction> = {
    type: 'success';
    result: Awaited<ReturnType<TFunc>>;
} | {
    type: 'error';
    error: any;
};
//# sourceMappingURL=tryCatchFinally.d.ts.map