"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createTryCatchFinally = void 0;
function createTryCatchFinally(func, interceptors) {
    let funcWrapper = function tryCatchFinallyWrapper(...funcArgs) {
        var _a;
        let isAsync = false;
        let funcRes;
        let ctx = { func, funcArgs, funcWrapper, funcThis: this };
        let tryRes = {};
        let onCatch = 'onCatch' in interceptors ? interceptors.onCatch : undefined;
        let onFinally = 'onFinally' in interceptors ? interceptors.onFinally : undefined;
        try {
            if (interceptors.onTry)
                tryRes = interceptors.onTry(ctx);
            if (tryRes.context)
                Object.assign(ctx, tryRes.context);
            if ('onCatch' in tryRes)
                onCatch = tryRes.onCatch;
            if ('onFinally' in tryRes)
                onFinally = tryRes.onFinally;
            funcRes = (_a = ctx.func) === null || _a === void 0 ? void 0 : _a.apply(ctx.funcThis, ctx.funcArgs);
            isAsync = isPromise(funcRes);
            ctx.funcOutcome = { type: 'success', result: funcRes };
            if (!isAsync) {
                return funcRes;
            }
            else
                return (function tryCatchFinallyPromiseWrapper() {
                    return __awaiter(this, void 0, void 0, function* () {
                        try {
                            ctx.funcOutcome = { type: 'success', result: yield funcRes };
                        }
                        catch (err) {
                            ctx.funcOutcome = { type: 'error', error: err };
                            onCatch === null || onCatch === void 0 ? void 0 : onCatch(ctx);
                        }
                        finally {
                            onFinally === null || onFinally === void 0 ? void 0 : onFinally(ctx);
                            if (ctx.funcOutcome) {
                                if (ctx.funcOutcome.type === 'success') {
                                    return ctx.funcOutcome.result;
                                }
                                else if (ctx.funcOutcome.type === 'error') {
                                    throw ctx.funcOutcome.error;
                                }
                            }
                        }
                    });
                })();
        }
        catch (error) {
            ctx.funcOutcome = { type: 'error', error };
            onCatch === null || onCatch === void 0 ? void 0 : onCatch(ctx);
        }
        finally {
            if (!isAsync) {
                onFinally === null || onFinally === void 0 ? void 0 : onFinally(ctx);
                if (ctx.funcOutcome) {
                    if (ctx.funcOutcome.type === 'success') {
                        return ctx.funcOutcome.result;
                    }
                    else if (ctx.funcOutcome.type === 'error') {
                        throw ctx.funcOutcome.error;
                    }
                }
            }
        }
    };
    Object.defineProperty(funcWrapper, 'name', { value: funcWrapper.name + '_' + (func.name || 'anonymous') });
    return funcWrapper;
}
exports.createTryCatchFinally = createTryCatchFinally;
function isPromise(val) {
    return val && (val instanceof Promise || typeof val.then === 'function');
}
